@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2015 Jan Marvin Garbuszus and Sebastian Jeworutzki
+ * Copyright (C) 2014-2019 Jan Marvin Garbuszus and Sebastian Jeworutzki
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -16,11 +16,15 @@
  */
 
 #include "readstata.h"
+#include "read_data.h"
 
 using namespace Rcpp;
 using namespace std;
 
-List read_dta(FILE * file, const bool missing) {
+List read_dta(FILE * file, const bool missing, const IntegerVector selectrows,
+              const CharacterVector selectcols,
+              const bool strlexport, const CharacterVector strlpath)
+{
   // stata_dta><header>
   test("stata_dta><header>", file);
   test("<release>", file);
@@ -30,7 +34,7 @@ List read_dta(FILE * file, const bool missing) {
   */
 
   int8_t fversion = 117L; //f = first
-  int8_t lversion = 118L; //l = last
+  int8_t lversion = 119L; //l = last
 
   std::string version(3, '\0');
   readstring(version, file, version.size());
@@ -43,7 +47,7 @@ List read_dta(FILE * file, const bool missing) {
   // check the release version.
   if (release<fversion || release>lversion)
   {
-    Rcpp::warning("File version is %d.\nVersion: Not a version 13/14 dta-file", release);
+    warning("File version is %d.\nVersion: Not a version 13/14 dta-file", release);
     return -1;
   }
 
@@ -65,6 +69,7 @@ List read_dta(FILE * file, const bool missing) {
     lbllen = 33;
     break;
   case 118:
+  case 119:
     nvarnameslen = 129;
     nformatslen = 57;
     nvalLabelslen = 129;
@@ -79,7 +84,7 @@ List read_dta(FILE * file, const bool missing) {
   test("<byteorder>", file);
 
   /*
-  * byteorder is a 4 byte character e.g. "LSF". MSF referes to big-memory data.
+  * byteorder is a 4 byte character e.g. "LSF". MSF refers to big-endian.
   */
 
   std::string byteorder(3, '\0');
@@ -96,8 +101,11 @@ List read_dta(FILE * file, const bool missing) {
   * Number of Variables
   */
 
-  uint16_t k = 0;
-  k = readbin(k, file, swapit);
+  uint32_t k = 0;
+  if (release < 119)
+    k = readbin((uint16_t)k, file, swapit);
+  if (release == 119)
+    k = readbin(k, file, swapit);
 
   //</K>
   test("</K>", file);
@@ -107,19 +115,22 @@ List read_dta(FILE * file, const bool missing) {
   * Number of Observations
   */
 
-  int64_t n = 0;
+  uint64_t n = 0;
 
-  if(release==117) {
-    n = readbin((int32_t)n, file, swapit);
-  }
-  if (release ==118) {
+  if (release == 117)
+    n = readbin((uint32_t)n, file, swapit);
+  if ((release == 118) | (release == 119))
     n = readbin(n, file, swapit);
-  }
 
   //</N>
   test("</N>", file);
   test("<label>", file);
 
+  // dim to return original dim for partial read files
+  IntegerVector dim(2);
+  dim(0) = n;
+  dim(1) = k;
+
   /*
   * A dataset may have a label e.g. "Written by R".
   * First we read its length (ndlabel), later the actual label (datalabel).
@@ -129,10 +140,10 @@ List read_dta(FILE * file, const bool missing) {
 
   uint16_t ndlabel = 0;
 
-  if (release==118)
-    ndlabel = readbin(ndlabel, file, swapit);
-  if (release==117)
+  if (release == 117)
     ndlabel = readbin((int8_t)ndlabel, file, swapit);
+  if ((release == 118) | (release == 119))
+    ndlabel = readbin(ndlabel, file, swapit);
 
   std::string datalabel(ndlabel, '\0');
 
@@ -217,7 +228,7 @@ List read_dta(FILE * file, const bool missing) {
   */
 
   IntegerVector vartype(k);
-  for (uint16_t i=0; i<k; ++i)
+  for (uint32_t i=0; i<k; ++i)
   {
     uint16_t nvartype = 0;
     nvartype = readbin(nvartype, file, swapit);
@@ -235,7 +246,7 @@ List read_dta(FILE * file, const bool missing) {
   std::string nvarnames(nvarnameslen, '\0');
 
   CharacterVector varnames(k);
-  for (uint16_t i=0; i<k; ++i)
+  for (uint32_t i=0; i<k; ++i)
   {
     readstring(nvarnames, file, nvarnames.size());
     varnames[i] = nvarnames;
@@ -247,18 +258,23 @@ List read_dta(FILE * file, const bool missing) {
 
   /*
   * sortlist. Stata stores the information which variable of a dataset was
-  * sorted. Depending on byteorder sortlist is written different. Currently we
+  * sorted. Depending on byteorder sortlist is written differently. Currently we
   * do not use this information.
   * Vector size is k+1.
   */
 
-  uint32_t big_k = k+1;
+  uint64_t big_k = k+1;
 
   IntegerVector sortlist(big_k);
-  for (uint32_t i=0; i<big_k; ++i)
+  for (uint64_t i=0; i<big_k; ++i)
   {
-    uint16_t nsortlist = 0;
-    nsortlist = readbin(nsortlist, file, swapit);
+    uint32_t nsortlist = 0;
+    
+    if ((release == 117) | (release == 118))
+      nsortlist = readbin((uint16_t)nsortlist, file, swapit);
+    if (release == 119)
+      nsortlist = readbin(nsortlist, file, swapit);
+    
     sortlist[i] = nsortlist;
   }
 
@@ -274,7 +290,7 @@ List read_dta(FILE * file, const bool missing) {
   std::string nformats(nformatslen, '\0');
 
   CharacterVector formats(k);
-  for (uint16_t i=0; i<k; ++i)
+  for (uint32_t i=0; i<k; ++i)
   {
     readstring(nformats, file, nformats.size());
     formats[i] = nformats;
@@ -293,7 +309,7 @@ List read_dta(FILE * file, const bool missing) {
   std::string nvalLabels(nvalLabelslen, '\0');
 
   CharacterVector valLabels(k);
-  for (uint16_t i=0; i<k; ++i)
+  for (uint32_t i=0; i<k; ++i)
   {
     readstring(nvalLabels, file, nvalLabels.size());
     valLabels[i] = nvalLabels;
@@ -310,7 +326,7 @@ List read_dta(FILE * file, const bool missing) {
   std::string nvarLabels (nvarLabelslen, '\0');
 
   CharacterVector varLabels(k);
-  for (uint16_t i=0; i<k; ++i)
+  for (uint32_t i=0; i<k; ++i)
   {
     readstring(nvarLabels, file, nvarLabels.size());
     varLabels[i] = nvarLabels;
@@ -324,7 +340,7 @@ List read_dta(FILE * file, const bool missing) {
   * characteristics. Stata can store additional information this way. It may
   * contain notes (for the dataset or a variable) or about label language sets.
   * Characteristics are not documented. We export them as attribute:
-  * expansion.fields. Characteristics are seperated by <ch> tags. Each <ch> has:
+  * expansion.fields. Characteristics are separated by <ch> tags. Each <ch> has:
   * nocharacter:  length of the characteristics
   * chvarname:    varname (binary 0 terminated)
   * chcharact:    characteristicsname (binary 0 terminated)
@@ -376,171 +392,84 @@ List read_dta(FILE * file, const bool missing) {
   * data. First a list is created with vectors. The vector type is defined by
   * vartype. Stata stores data columnwise so we loop over it and store the
   * data in the list of the first step. Third variable- and row-names are
-  * attatched and the list type is changed to data.frame.
+  * attached and the list type is changed to data.frame.
   */
 
-  // 1. create the list
-  List df(k);
-  for (uint16_t i=0; i<k; ++i)
-  {
-    int const type = vartype[i];
-    switch(type)
-    {
-    case 65526:
-    case 65527:
-      SET_VECTOR_ELT(df, i, NumericVector(no_init(n)));
-      break;
+  uint64_t nmin = selectrows(0), nmax = selectrows(1);
+  uint64_t nn   = 0;
 
-    case 65528:
-    case 65529:
-    case 65530:
-      SET_VECTOR_ELT(df, i, IntegerVector(no_init(n)));
-      break;
-
-    default:
-      SET_VECTOR_ELT(df, i, CharacterVector(no_init(n)));
-    break;
-    }
+  // if  selectrows is c(0,0) use full data
+  if ((nmin == 0) && (nmax == 0)){
+    nmin = 1;
+    nmax = n;
   }
 
+  // make sure that n is not greater than nmax or nmin
+  if (n < nmax)
+    nmax = n;
+  if (n < nmin)
+    nmin = n;
+
+  // sequences of column and row
+  IntegerVector cvec = seq(0, (k-1));
+  IntegerVector rvec = seq(nmin, nmax);
+  nn = rvec.size();
+
+  // use c indexing starting at 0
+  nmin = nmin -1;
+  nmax = nmax -1;
+
+  // calculate length of each variable stored in file. Calculate row length
+  IntegerVector rlen = calc_rowlength(vartype);
+  uint64_t rlength = sum(rlen);
+
+  // check if vars are selected
+  std::string selcols = as<std::string>(selectcols(0));
+  bool selectvars = selcols != "";
+
+  // select vars: either select every var or only matched cases. This will
+  // return index positions of the selected variables. If non are selected the
+  // index position is cvec
+  IntegerVector select = cvec, nselect;
+  if (selectvars)
+    select = choose(selectcols, varnames);
+
+  // separate the selected from the not selected cases
+  LogicalVector ll = is_na(select);
+  nselect = cvec[ll == 1];
+  select = cvec[ll == 0];
+
+  uint32_t kk = select.size();
+
+  // shrink variables to selected size
+  CharacterVector varnames_kk = varnames[select];
+  IntegerVector vartype_kk = vartype[select];
+  IntegerVector vartype_s = vartype;
+
+  // replace not selected cases with their negative size values
+  IntegerVector rlen2 = rlen[nselect];
+  rlen2 = -rlen2;
+  vartype_s[nselect] = rlen2;
+
+
+
+  // Use vartype_s to calculate jump
+  IntegerVector vartype_sj = calc_jump(vartype_s);
+
   // 2. fill it with data
-  for(uint32_t j=0; j<n; ++j)
-  {
-    for (uint16_t i=0; i<k; ++i)
-    {
-      int const type = vartype[i];
-      switch(type < 2046 ? 2045 : type)
-      {
-        // double
-      case 65526:
-      {
-        double val_d = 0;
-        val_d = readbin(val_d, file, swapit);
 
-        if ((missing == 0) && !(val_d == R_NegInf) && ((val_d<STATA_DOUBLE_NA_MIN) || (val_d>STATA_DOUBLE_NA_MAX)) )
-          REAL(VECTOR_ELT(df,i))[j] = NA_REAL;
-        else
-          REAL(VECTOR_ELT(df,i))[j] = val_d;
-        break;
-      }
-        // float
-      case 65527:
-      {
-        float val_f = 0;
-        val_f = readbin(val_f, file, swapit);
+  // skip into the data part
+  fseeko64(file, rlength * nmin, SEEK_CUR);
 
-        if ((missing == 0) && ((val_f<STATA_FLOAT_NA_MIN) || (val_f>STATA_FLOAT_NA_MAX)) )
-          REAL(VECTOR_ELT(df,i))[j] = NA_REAL;
-        else
-          REAL(VECTOR_ELT(df,i))[j] = val_f;
-        break;
-      }
-        //long
-      case 65528:
-      {
-        int32_t val_l = 0;
-        val_l = readbin(val_l, file, swapit);
+  List df = read_data(file, vartype_kk, missing, release, nn, kk,
+                      vartype_sj, byteorder, swapit);
 
-        if ((missing == 0) && ((val_l<STATA_INT_NA_MIN) || (val_l>STATA_INT_NA_MAX)) )
-          INTEGER(VECTOR_ELT(df,i))[j]  = NA_INTEGER;
-        else
-          INTEGER(VECTOR_ELT(df,i))[j] = val_l;
-        break;
-      }
-        // int
-      case 65529:
-      {
-        int16_t val_i = 0;
-        val_i = readbin(val_i, file, swapit);
-
-        if ((missing == 0) && ((val_i<STATA_SHORTINT_NA_MIN) || (val_i>STATA_SHORTINT_NA_MAX)) )
-          INTEGER(VECTOR_ELT(df,i))[j] = NA_INTEGER;
-        else
-          INTEGER(VECTOR_ELT(df,i))[j] = val_i;
-        break;
-      }
-        // byte
-      case 65530:
-      {
-        int8_t val_b = 0;
-        val_b = readbin(val_b, file, swapit);
-
-        if (missing == 0 && ( (val_b<STATA_BYTE_NA_MIN) || (val_b>STATA_BYTE_NA_MAX)) )
-          INTEGER(VECTOR_ELT(df,i))[j] = NA_INTEGER;
-        else
-          INTEGER(VECTOR_ELT(df,i))[j] = val_b;
-        break;
-      }
-        // strings with 2045 or fewer characters
-      case 2045:
-      {
-        int32_t len = 0;
-        len = vartype[i];
-        std::string val_s (len, '\0');
-
-        readstring(val_s, file, val_s.size());
-        as<CharacterVector>(df[i])[j] = val_s;
-        break;
-      }
-        // string of any length
-      case 32768:
-      {// strL 2*4bit or 2 + 6 bit
-        //char val_strl[22];
-
-        // FixMe: Strl in 118
-        switch (release)
-        {
-
-        case 117:
-          {
-            uint32_t v = 0, o = 0;
-
-            v = readbin(v, file, swapit);
-            o = readbin(o, file, swapit);
-
-            stringstream val_stream;
-            val_stream << v << '_' << o;
-            string val_strl = val_stream.str();
-            //sprintf(val_strl, "%010d%010d", v, o);
-            as<CharacterVector>(df[i])[j] = val_strl;
-            break;
-          }
-        case 118:
-          {
-            int16_t v = 0;
-            int64_t o = 0, z = 0;
-
-            z = readbin(z, file, swapit);
-
-            // works for LSF on little- and big-endian
-            if(byteorder.compare("LSF")==0) {
-              v = (int16_t)z;
-              o = (z >> 16);
-            }
-
-            // works if we read a big-endian file on little-endian
-            if(byteorder.compare("MSF")==0) {
-              v = (z >> 48) & ((1 << 16) - 1);
-              o = z & ((1 << 16) - 1);
-            }
-
-            stringstream val_stream;
-            val_stream << v << '_' << o;
-            string val_strl = val_stream.str();
-
-            as<CharacterVector>(df[i])[j] = val_strl;
-            break;
-          }
-        }
-      }
-      }
-    }
-  }
+  // skip to end of data part
+  fseeko64(file, rlength * (n - nmax -1), SEEK_CUR);
 
   // 3. Create a data.frame
-  R_xlen_t nrows = Rf_length(df[0]);
-  df.attr("row.names") = IntegerVector::create(NA_INTEGER, nrows);
-  df.attr("names") = varnames;
+  df.attr("row.names") = rvec;
+  df.attr("names") = varnames_kk;
   df.attr("class") = "data.frame";
 
   //</data>
@@ -548,8 +477,8 @@ List read_dta(FILE * file, const bool missing) {
   test("<strls>", file);
 
   /*
-  * strL. Stata 13 introduced long strings up to 2 billon characters. strLs are
-  * sperated by "GSO".
+  * strL. Stata 13 introduced long strings up to 2 billion characters. strLs are
+  * separated by "GSO".
   * (v,o): Position in the data.frame.
   * t:     129/130 defines whether or not the strL is stored with a binary 0.
   * len:   length of the strL.
@@ -565,7 +494,7 @@ List read_dta(FILE * file, const bool missing) {
   CharacterVector strlvalues(0);
   CharacterVector strlnames(0);
 
-  while(gso.compare(tags)==0)
+  while (gso.compare(tags)==0)
   {
     CharacterVector strls(2);
     string ref;
@@ -583,39 +512,56 @@ List read_dta(FILE * file, const bool missing) {
       stringstream val_stream;
       val_stream << v << '_' << o;
       ref.assign(val_stream.str());
-      //sprintf(ref, "%010d%010d", v, o);
+
       break;
     }
     case 118:
+    case 119:
     {
       uint32_t v = 0;
       uint64_t o = 0;
-      // uint64_t z = 0;
+
       v = readbin(v, file, swapit);
       o = readbin(o, file, swapit);
-      // z = readbin(z, file, swapit);
 
       stringstream val_stream;
       val_stream << v << '_' << o;
       ref.assign(val_stream.str());
-      //sprintf(ref, "%010d%010ld", v, o);
 
       break;
     }
     }
 
-    // (129 = binary) | (130 = ascii)
+    // (129 = binary) | (130 = ascii) Note:
+    // if 130 full len contains the string. if 130 len includes trailing \0.
+    // that does not affect us. we read the full len, and if \0 occurs R
+    // will print only the string up to that position. we write 129
     uint8_t t = 0;
     t = readbin(t, file, swapit);
 
     uint32_t len = 0;
     len = readbin(len, file, swapit);
 
-    // 129 len = len; 130 len = len +'\0';
-
     std::string strl(len, '\0');
+
     readstring(strl, file, strl.size());
 
+    // write strl to file. Stata allows binary files in strls
+    if (strlexport) {
+
+      std::string path = Rcpp::as<std::string>(strlpath);
+      std::string outputpath = path + "/" + ref;
+
+      ofstream file1(outputpath.c_str(), ios::out | ios::binary);
+      if (file1.good()) {
+        file1.write(strl.c_str(), strl.size());
+        file1.close();
+      } else {
+        Rcpp::Rcout << "strl export failed" << std::endl;
+      }
+
+    }
+
     strlvalues.push_back( strl );
     strlnames.push_back( ref );
 
@@ -631,7 +577,7 @@ List read_dta(FILE * file, const bool missing) {
   test("<value_labels>", file);
 
   /*
-  * labels are seperated by <lbl>-tags. Labels may appear in any order e.g.
+  * labels are separated by <lbl>-tags. Labels may appear in any order e.g.
   * 2 "female" 1 "male 9 "missing". They are stored as tables.
   * nlen:     length of label.
   * nlabname: label name.
@@ -647,7 +593,7 @@ List read_dta(FILE * file, const bool missing) {
 
   List labelList = List(); //put labels into this list
 
-  while(lbltag.compare(tag)==0)
+  while (lbltag.compare(tag)==0)
   {
     int32_t nlen = 0, labn = 0, txtlen = 0, noff = 0, val = 0;
 
@@ -738,16 +684,18 @@ List read_dta(FILE * file, const bool missing) {
   test("ue_labels>", file);
   test("</stata_dta>", file);
 
-
-
   /*
    * assign attributes to the resulting data.frame
    */
 
+  formats = formats[select];
+  valLabels = valLabels[select];
+  varLabels = varLabels[select];
+
   df.attr("datalabel") = datalabelCV;
   df.attr("time.stamp") = timestampCV;
   df.attr("formats") = formats;
-  df.attr("types") = vartype;
+  df.attr("types") = vartype_kk;
   df.attr("val.labels") = valLabels;
   df.attr("var.labels") = varLabels;
   df.attr("version") = versionIV;
@@ -755,6 +703,7 @@ List read_dta(FILE * file, const bool missing) {
   df.attr("expansion.fields") = ch;
   df.attr("strl") = strlvalues;
   df.attr("byteorder") = wrap(byteorder);
+  df.attr("orig.dim") = dim;
 
   return df;
 }
